#include "ns3/command-line.h"
#include "ns3/config.h"
#include "ns3/double.h"
#include "ns3/internet-stack-helper.h"
#include "ns3/ipv4-address-helper.h"
#include "ns3/log.h"
#include "ns3/mobility-helper.h"
#include "ns3/mobility-model.h"
#include "ns3/ssid.h"
#include "ns3/string.h"
#include "ns3/yans-wifi-channel.h"
#include "ns3/yans-wifi-helper.h"
#include "ns3/flow-monitor-module.h"
#include "ns3/flow-monitor-helper.h"
#include "ns3/netanim-module.h"
using namespace ns3;

NS_LOG_COMPONENT_DEFINE("WifiSimpleInfra");
void
ReceivePacket(Ptr<Socket> socket)
{
	
    while (socket->Recv())
    {
    
    }
 
}

static void
GenerateTraffic(Ptr<Socket> socket, uint32_t pktSize, uint32_t pktCount, Time pktInterval)
{
    if (pktCount > 0)
    {
        NS_LOG_INFO("Generating one packet of size " << pktSize);
        socket->Send(Create<Packet>(pktSize));
        Simulator::Schedule(pktInterval,
                            &GenerateTraffic,
                            socket,
                            pktSize,
                            pktCount - 1,
                            pktInterval);
    }
    else
    {
        socket->Close();
    }
}

int
main(int argc, char* argv[])
{
    std::string phyMode("DsssRate1Mbps");
    double rss = -80;           // -dBm
    uint32_t packetSize = 1000; // bytes
    uint32_t numPackets = 100;
    
    Time interval = Seconds(0.003);
    bool verbose = false;
    
    uint32_t sinkNode = 1;
    uint32_t sourceNode = 0;
    
     uint32_t sinkNode1 = 1;
    uint32_t sourceNode1 = 2;
    
    uint32_t sinkNode2 = 0;
    uint32_t sourceNode2 =2 ;
    
    uint32_t packetSize1 = 1200; // bytes
    uint32_t numPackets1 = 200;
    
    uint32_t packetSize2 = 1700; // bytes
    uint32_t numPackets2 = 300;

    CommandLine cmd(__FILE__);
    cmd.AddValue("phyMode", "Wifi Phy mode", phyMode);
    cmd.AddValue("rss", "received signal strength", rss);
    cmd.AddValue("packetSize", "size of application packet sent", packetSize);
    cmd.AddValue("numPackets", "number of packets generated", numPackets);
    cmd.AddValue("interval", "interval between packets", interval);
    cmd.AddValue("verbose", "turn on all WifiNetDevice log components", verbose);
    cmd.Parse(argc, argv);

    // Fix non-unicast data rate to be the same as that of unicast
    Config::SetDefault("ns3::WifiRemoteStationManager::NonUnicastMode", StringValue(phyMode));
    
    NS_LOG_UNCOND ("Topology");
    NS_LOG_UNCOND ("    n3 (Station  node) ");
    NS_LOG_UNCOND ("    . ");
    NS_LOG_UNCOND ("    . ");
    NS_LOG_UNCOND ("    n1 (AP node)");
    NS_LOG_UNCOND ("   .  .   ");
    NS_LOG_UNCOND ("  .    .   ");
    NS_LOG_UNCOND (" n0    n2  (Station nodes)");
    
    NS_LOG_UNCOND("-----------------------------------------------------------");
    NodeContainer c;
    c.Create(4);

    // The below set of helpers will help us to put together the wifi NICs we want
    WifiHelper wifi;
    if (verbose)
    {
        WifiHelper::EnableLogComponents(); // Turn on all Wifi logging
    }
    wifi.SetStandard(WIFI_STANDARD_80211b);

    YansWifiPhyHelper wifiPhy;
    // This is one parameter that matters when using FixedRssLossModel
    // set it to zero; otherwise, gain will be added
    wifiPhy.Set("RxGain", DoubleValue(0)); //Include about RxGain in the additional reading material
    // ns-3 supports RadioTap and Prism tracing extensions for 802.11b
    wifiPhy.SetPcapDataLinkType(WifiPhyHelper::DLT_IEEE802_11_RADIO);

    YansWifiChannelHelper wifiChannel;
    wifiChannel.SetPropagationDelay("ns3::ConstantSpeedPropagationDelayModel");
    // The below FixedRssLossModel will cause the rss to be fixed regardless
    // of the distance between the two stations, and the transmit power
    wifiChannel.AddPropagationLoss("ns3::FixedRssLossModel", "Rss", DoubleValue(rss));
    wifiPhy.SetChannel(wifiChannel.Create());

    // Add a mac and disable rate control
    WifiMacHelper wifiMac;
    wifi.SetRemoteStationManager("ns3::ConstantRateWifiManager",
                                 "DataMode",
                                 StringValue(phyMode),
                                 "ControlMode",
                                 StringValue(phyMode));

    // Setup the rest of the MAC
    Ssid ssid = Ssid("wifi-default");
    // setup STA
    wifiMac.SetType("ns3::StaWifiMac", "Ssid", SsidValue(ssid));
    NetDeviceContainer staDevice = wifi.Install(wifiPhy, wifiMac, c);
    NetDeviceContainer devices = staDevice;
   
    // setup AP
    wifiMac.SetType("ns3::ApWifiMac", "Ssid", SsidValue(ssid));
    NetDeviceContainer apDevice = wifi.Install(wifiPhy, wifiMac, c.Get(1));
    devices.Add(apDevice);

    // Note that with FixedRssLossModel, the positions below are not
    // used for received signal strength.
    MobilityHelper mobility;
    Ptr<ListPositionAllocator> positionAlloc = CreateObject<ListPositionAllocator>();
    positionAlloc->Add(Vector(3.0, 4.0, 3.0));
    positionAlloc->Add(Vector(5.0, 4.0, 3.0));
    positionAlloc->Add(Vector(7.0,4.0,3.0));
    positionAlloc->Add(Vector(5.0,5.0,4.0));
    mobility.SetPositionAllocator(positionAlloc);
    mobility.SetMobilityModel("ns3::ConstantPositionMobilityModel");
    mobility.Install(c);

    InternetStackHelper internet;
    internet.Install(c);



    Ipv4AddressHelper ipv4;
    ipv4.SetBase("10.1.1.0", "255.255.255.0");
    Ipv4InterfaceContainer i = ipv4.Assign(devices);


//udp flow-1:

    TypeId tid = TypeId::LookupByName("ns3::UdpSocketFactory");
    Ptr<Socket> recvSink = Socket::CreateSocket(c.Get(sinkNode), tid);
    InetSocketAddress local = InetSocketAddress(Ipv4Address::GetAny(), 80);
    recvSink->Bind(local);
    recvSink->SetRecvCallback(MakeCallback(&ReceivePacket));

    Ptr<Socket> source = Socket::CreateSocket(c.Get(sourceNode), tid);
    InetSocketAddress remote = InetSocketAddress(i.GetAddress(sinkNode, 0), 80);
    source->Connect(remote);

//udp flow-2:
    TypeId tid1 = TypeId::LookupByName("ns3::UdpSocketFactory");
    Ptr<Socket> recvSink1 = Socket::CreateSocket(c.Get(sinkNode1), tid1);
    InetSocketAddress local1 = InetSocketAddress(Ipv4Address::GetAny(), 82);
    recvSink1->Bind(local1);
    recvSink1->SetRecvCallback(MakeCallback(&ReceivePacket));

    Ptr<Socket> source1 = Socket::CreateSocket(c.Get(sourceNode1), tid1);
    InetSocketAddress remote1 = InetSocketAddress(i.GetAddress(sinkNode1, 0), 82);
    source1->Connect(remote1);
//udp flow-3:
   TypeId tid2 = TypeId::LookupByName("ns3::UdpSocketFactory");
    Ptr<Socket> recvSink2 = Socket::CreateSocket(c.Get(sinkNode2), tid2);
    InetSocketAddress local2 = InetSocketAddress(Ipv4Address::GetAny(), 91);
    recvSink2->Bind(local2);
    recvSink2->SetRecvCallback(MakeCallback(&ReceivePacket));

    Ptr<Socket> source2 = Socket::CreateSocket(c.Get(sourceNode2), tid2);
    InetSocketAddress remote2 = InetSocketAddress(i.GetAddress(sinkNode2, 0), 91);
    source2->Connect(remote2);   
    
    // Tracing
    wifiPhy.EnablePcap("wifi-simple-infra", devices);

    // Output what we are doing
    std::cout << "Testing " << numPackets << " packets sent with receiver rss " << rss << std::endl;

    Simulator::Schedule(Seconds(10.0),&GenerateTraffic,source,packetSize,numPackets,interval);

    Simulator::Schedule(
                                   Seconds(10.0),
                                   &GenerateTraffic,
                                   source1,
                                   packetSize1,
                                   numPackets1,
                                   interval);
      Simulator::Schedule(
                                   Seconds(10.0),
                                   &GenerateTraffic,
                                   source2,
                                   packetSize2,
                                   numPackets2,
                                   interval);
      std::cout << "Testing " << numPackets1 << " packets sent with receiver rss " << rss << std::endl;
      std::cout << "Testing " << numPackets2 << " packets sent with receiver rss " << rss << std::endl;
   
     
     Ptr<FlowMonitor> flowMonitor;
     FlowMonitorHelper flowHelper;
     flowMonitor= flowHelper.InstallAll();
     Simulator::Stop(Seconds(500.0));
     AnimationInterface anim ("animationwifi-infra-2.xml");
     NS_LOG_INFO ("Run Simulation.");
     flowMonitor->SerializeToXmlFile("wifi-infra-final.xml", true, true);
     
    
     Simulator::Run();
    
    Ptr<Ipv4FlowClassifier> classifier=DynamicCast<Ipv4FlowClassifier> (flowHelper.GetClassifier());
    std::map<FlowId,FlowMonitor::FlowStats> stats=flowMonitor->GetFlowStats();
     

    for(std::map<FlowId,FlowMonitor::FlowStats>::const_iterator iter=stats.begin();iter!=stats.end();++iter)
    {
    Ipv4FlowClassifier::FiveTuple t =classifier->FindFlow(iter->first);
    NS_LOG_UNCOND("Flow ID:- "<< iter->first<< " Source addr: "<< t.sourceAddress << " Dest Addr: "<<t.destinationAddress ); 
     uint32_t protocol = t.protocol;
    if (protocol == 6) {
        std::cout << "Type: TCP Flow" << std::endl;
    } else if (protocol == 17) {
        std::cout << "Type: UDP Flow" << std::endl;
    } 
    	  NS_LOG_UNCOND("Tx Packets = " << iter->second.txPackets);
    	  NS_LOG_UNCOND("Rx Packets = " << iter->second.rxPackets);
    	  NS_LOG_UNCOND("Lost packets = " << iter->second.lostPackets);
    	  NS_LOG_UNCOND("Times forwarded = " << iter->second.timesForwarded);
    	  
    	  
    	  NS_LOG_UNCOND("Delay = " << iter->second.delaySum);
    	  NS_LOG_UNCOND("Throughput: " << iter->second.rxBytes * 8.0 / (iter->second.timeLastRxPacket.GetSeconds()-iter->second.timeFirstTxPacket.GetSeconds()) / 1024  << " Kbps");
    	  NS_LOG_UNCOND("-----------------------------------------------------------");
   
    }
    
    Simulator::Destroy();
    return 0;
}
